@@ -52,9 +52,12 @@ class Connection(object):
         if loop.get_debug():
             self._source_traceback = traceback.extract_stack(sys._getframe(1))
 
+    def __repr__(self):
+        return 'Connection<{}>'.format(self._key)
+
     def __del__(self, _warnings=warnings):
         if self._transport is not None:
-            _warnings.warn("Unclosed connection {!r}".format(self),
+            _warnings.warn('Unclosed connection {!r}'.format(self),
                            ResourceWarning)
             if hasattr(self._loop, 'is_closed'):
                 if self._loop.is_closed():
@@ -146,11 +149,14 @@ class BaseConnector(object):
         if not self._conns:
             return
 
+        conns = [repr(c) for c in self._conns.values()]
+
         self.close()
 
         _warnings.warn("Unclosed connector {!r}".format(self),
                        ResourceWarning)
         context = {'connector': self,
+                   'connections': conns,
                    'message': 'Unclosed connector'}
         if self._source_traceback is not None:
             context['source_traceback'] = self._source_traceback
@@ -216,14 +222,16 @@ class BaseConnector(object):
 
     def close(self):
         """Close all opened transports."""
+        ret = asyncio.Future(loop=self._loop)
+        ret.set_result(None)
         if self._closed:
-            return
+            return ret
         self._closed = True
 
         try:
             if hasattr(self._loop, 'is_closed'):
                 if self._loop.is_closed():
-                    return
+                    return ret
 
             for key, data in self._conns.items():
                 for transport, proto, t0 in data:
@@ -239,6 +247,7 @@ class BaseConnector(object):
             self._conns.clear()
             self._acquired.clear()
             self._cleanup_handle = None
+        return ret
 
     @property
     def closed(self):
@@ -271,12 +280,24 @@ class BaseConnector(object):
         """Get from pool or create new connection."""
         key = (req.host, req.port, req.ssl)
 
-        # use short-circuit
-        if self._limit is not None:
-            while len(self._acquired[key]) >= self._limit:
-                fut = asyncio.Future(loop=self._loop)
-                self._waiters[key].append(fut)
-                yield from fut
+        limit = self._limit
+        if limit is not None:
+            fut = asyncio.Future(loop=self._loop)
+            waiters = self._waiters[key]
+
+            # The limit defines the maximum number of concurrent connections
+            # for a key. Waiters must be counted against the limit, even before
+            # the underlying connection is created.
+            available = limit - len(waiters) - len(self._acquired[key])
+
+            # Don't wait if there are connections available.
+            if available > 0:
+                fut.set_result(None)
+
+            # This connection will now count towards the limit.
+            waiters.append(fut)
+
+            yield from fut
 
         transport, proto = self._get(key)
         if transport is None:
@@ -369,6 +390,8 @@ class BaseConnector(object):
         raise NotImplementedError()
 
 
+_SSL_OP_NO_COMPRESSION = getattr(ssl, "OP_NO_COMPRESSION", 0)
+
 _marker = object()
 
 
@@ -382,13 +405,14 @@ class TCPConnector(BaseConnector):
         https://en.wikipedia.org/wiki/Transport_Layer_Security#Certificate_pinning
     :param bool resolve: Set to True to do DNS lookup for host name.
     :param family: socket address family
+    :param local_addr: local :class:`tuple` of (host, port) to bind socket to
     :param args: see :class:`BaseConnector`
     :param kwargs: see :class:`BaseConnector`
     """
 
     def __init__(self, *, verify_ssl=True, fingerprint=None,
                  resolve=_marker, use_dns_cache=_marker,
-                 family=0, ssl_context=None,
+                 family=0, ssl_context=None, local_addr=None,
                  **kwargs):
         super().__init__(**kwargs)
 
@@ -427,6 +451,7 @@ class TCPConnector(BaseConnector):
         self._cached_hosts = {}
         self._ssl_context = ssl_context
         self._family = family
+        self._local_addr = local_addr
 
     @property
     def verify_ssl(self):
@@ -449,7 +474,7 @@ class TCPConnector(BaseConnector):
                 sslcontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                 sslcontext.options |= ssl.OP_NO_SSLv2
                 sslcontext.options |= ssl.OP_NO_SSLv3
-                sslcontext.options |= ssl.OP_NO_COMPRESSION
+                sslcontext.options |= _SSL_OP_NO_COMPRESSION
                 sslcontext.set_default_verify_paths()
             else:
                 sslcontext = ssl.create_default_context()
@@ -552,7 +577,8 @@ class TCPConnector(BaseConnector):
                     self._factory, host, port,
                     ssl=sslcontext, family=hinfo['family'],
                     proto=hinfo['proto'], flags=hinfo['flags'],
-                    server_hostname=hinfo['hostname'] if sslcontext else None)
+                    server_hostname=hinfo['hostname'] if sslcontext else None,
+                    local_addr=self._local_addr)
                 has_cert = transp.get_extra_info('sslcontext')
                 if has_cert and self._fingerprint:
                     sock = transp.get_extra_info('socket')
@@ -637,7 +663,10 @@ class ProxyConnector(TCPConnector):
         if hdrs.AUTHORIZATION in proxy_req.headers:
             auth = proxy_req.headers[hdrs.AUTHORIZATION]
             del proxy_req.headers[hdrs.AUTHORIZATION]
-            req.headers[hdrs.PROXY_AUTHORIZATION] = auth
+            if not req.ssl:
+                req.headers[hdrs.PROXY_AUTHORIZATION] = auth
+            else:
+                proxy_req.headers[hdrs.PROXY_AUTHORIZATION] = auth
 
         if req.ssl:
             # For HTTPS requests over HTTP proxy
@@ -674,6 +703,8 @@ class ProxyConnector(TCPConnector):
                 transport, proto = yield from self._loop.create_connection(
                     self._factory, ssl=self.ssl_context, sock=rawsock,
                     server_hostname=req.host)
+            finally:
+                proxy_resp.close()
 
         return transport, proto
 
